@@ -2,9 +2,10 @@ from contextlib import closing
 from datetime import date
 from distutils.version import StrictVersion
 import logging
+import math
 
 from column import *
-from exceptions import classify_error
+from exceptions import classify_error, BrokenPipeError
 from settings import settings
 from utils import _camel_to_snake, dump, escaped
 from warehouse import Warehouse
@@ -55,6 +56,12 @@ class _ColumnSet(object):
     def natural_keys(self):
         return [c for c in self.columns if isinstance(c, NaturalKey)]
 
+    @property
+    def composite_key(self):
+        """Returns the values which make up the composite unique key."""
+        _ = (AutoColumn, ApplicableFrom, HashKey, Metric)
+        return [c for c in self.columns if not isinstance(c, _)]
+
 
 class TableMetaclass(type):
     """ Metaclass for constructing all Table classes. This applies number
@@ -84,6 +91,8 @@ class TableMetaclass(type):
             cls.__metrics__ = column_set.metrics
         if "__naturalkeys__" in dir(cls):
             cls.__naturalkeys__ = column_set.natural_keys
+        if "__compositekey__" in dir(cls):
+            cls.__compositekey__ = column_set.composite_key
 
         return cls
 
@@ -110,7 +119,7 @@ class Table(object):
         "COLLATE": "utf8_bin",
     }
 
-    INSERT = "INSERT"
+    INSERT = "INSERT IGNORE"
 
     @classmethod
     def create_trigger(cls):
@@ -226,6 +235,14 @@ class Table(object):
                 source.finish(cls)
         else:
             raise NotImplementedError("No data source defined")
+    
+    @classmethod
+    def batch(cls, instances):
+        """ Subdivides instances into smaller batches ready for insertion."""
+        batch_size = settings.BATCH_SIZE
+        batch_number = int(math.ceil(len(instances) / float(batch_size)))
+        batches = [instances[i * batch_size:(i + 1) * batch_size] for i in xrange(batch_number)]
+        return batches
 
     @classmethod
     def insert(cls, *instances):
@@ -234,26 +251,51 @@ class Table(object):
         if instances:
             columns = [column for column in cls.__columns__
                        if not isinstance(column, AutoColumn)]
+
             sql = "%s INTO %s (\n  %s\n)\n" % (
                 cls.INSERT, escaped(cls.__tablename__),
                 ",\n  ".join(escaped(column.name) for column in columns))
-            link = "VALUES"
-            for instance in instances:
-                values = []
-                for column in columns:
-                    value = instance[column.name]
-                    values.append(dump(value))
-                sql += link + (" (\n  %s\n)" % ",\n  ".join(values))
-                link = ","
 
-            connection = Warehouse.get()
-            with closing(connection.cursor()) as cursor:
-                try:
-                    cursor.execute(sql)
-                except:
-                    connection.rollback()
-                else:
-                    connection.commit()
+            insert_statement = sql
+            link = "VALUES"
+
+            batches = cls.batch(instances)
+            for iteration, batch in enumerate(batches, start=1):
+                log.debug('Inserting batch %s' % (iteration),
+                          extra={"table": cls.__tablename__})
+
+                for instance in batch:
+                    values = []
+                    for column in columns:
+                        value = instance[column.name]
+                        values.append(dump(value))
+                    sql += link + (" (\n  %s\n)" % ",\n  ".join(values))
+                    link = ","
+
+                for i in range(1, 3):
+                    connection = Warehouse.get()
+                    try:
+                        cursor = connection.cursor()
+                        cursor.execute(sql)
+                        cursor.close()
+
+                    except Exception as e:
+                        classify_error(e)
+                        if e.__class__ == BrokenPipeError and i == 1:
+                            log.info(
+                                'Trying once more with a fresh connection',
+                                extra={"table": cls.__tablename__}
+                                )
+                            connection.close()
+                        else:
+                            log.error(e)
+                            return
+                    else:
+                        connection.commit()
+                        break
+
+        log.debug('Finished updating %s' % cls.__tablename__,
+                  extra={"table": cls.__tablename__})
 
     @classmethod
     def update(cls, since=None, historical=False):
